#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
gqrx_settrx.py

Simple control program setting frequency from gqrx
==================================================

(C) 2019,2020 DG7BBP, Jens Rosebrock

"""
_version = "1.16"

import six
import argparse
import collections
import json
import os
import serial
import socket
import sys
import time
import subprocess
import shlex
import readline

import threading


EOT = '\x04'
LF = '\x0A'
_verbose_level = -1
_f_offset = 0
_initial_split_offset = 287500000
# for old lnb _initial_split_offset = 287700000

_split_offset = _initial_split_offset
_finish = False
_target_beacon_freq = 144749000

def show_qo100(f):
    """
    """
    global _target_beacon_freq
    _qo_downlink = 10489750000
    _modes = [(0, 5000, "CW-Beacon"),
              (5000, 40000, "CW"),
              (40000, 80000, "NB Dig"),
              (80000, 150000, "WB Dig"),
              (150000, 245000, "SSB"),
              (245000, 250000, "Beacon"),
              (255000, 350000, "SSB"),
              (350000, 495000, "Mixed"),
              (495000, 500000, "ExpBeac")]
    lo_all = _qo_downlink - _target_beacon_freq 
    total_f = f - _target_beacon_freq
    offset_f = total_f + 250000
    # print "input f: %s" %f
    # print "offset_f : %s" % offset_f
    mode = "Outside transponder"
    for l, u, m in _modes:
        if offset_f >= l and offset_f < u:
            mode = m
            break
    real_f = (_qo_downlink + total_f) / 1e6
    six.print_("RX on : %f Mode: %s" % (real_f, mode))

def log(level, msg):
    global _verbose_level
    if level <= _verbose_level:
        sys.stdout.write("%s\n" % msg)


def show_error(msg):
    sys.stderr.write("%s\n" % msg)

def rinput(msg):
    if six.PY2:
        return raw_input(msg)
    else:
        return input(msg)

def show_exception(e):
    msg = e.message
    if not msg:
        msg = e.strerror
    if not msg:
        msg = "%s" % e
    show_error(msg)


class HamLibConnection(object):
    """
    Simple class for requesting frequency from rigctld
    """

    def __init__(self, host_port):
        self._host_port = host_port
        self._socket = None

    def __enter__(self):
        """
        Allow with statement
        """
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        """
        self.disconnect()

    def connect(self, retry=1):
        """
        Connect to rightld
        :raises socket.error
        """
        addr_data = self.url_to_host(self._host_port, 4532)
        if len(addr_data)==2:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        last_err = None
        for i in range(retry):
            last_err = None
            try:
                s.connect(addr_data)
                break
            except (socket.error,)  as e:
                last_err = e
                time.sleep(1)
        if last_err is  None:
            self._socket = s
        else:
            raise last_err

    def disconnect(self):
        """
        Disconnect from rigtld
        """
        if self._socket is not None:
            self._socket.close()

    def url_to_host(self, host_port, default_port):
        """
        get hostname, ip address from host_port string like:
        <hostname>[:<port>]
        <ipv4_addr>[:<port>]
        \[<ipv6_addr>\][:<port>]
        Port is optional
        """
        splitted = host_port.split(":")
        port_str = None
        is_ipv6 = False
        if len(splitted) == 1:
            # ip v4/hostname no port
            host = host_port
        elif len(splitted) == 2:
            # ip v4/hostname given port
            host = splitted[0]
            port_str = splitted[1]
        else:
            log(3, "url_to_host: IPV6 %s" % (splitted,))
            is_ipv6 = True
            # some kind of ip6
            if splitted[-1].endswith("]"):
                # no port
                log(3,"url_to_host: NO PORT")
                host = host_port
                if host.startswith("[") and host.endswith("]"):
                    host = host[1:-1]
            else:
                port = splitted[-1]
                host  = ":".join(splitted[:-1])
                if host.startswith("[") and host.endswith("]"):
                    host = host[1:-1]
        if port_str is None or not port_str.isdigit():
            port = default_port
        else:
            port = int(port_str)
        log(3, "url_to_host: HOST %s PORT: %s %s" % (host, port, is_ipv6))
        if is_ipv6:
            return host, port , 0, 0
        else:
            return host, port

    def read_frequency(self):
        """
        Simple read function for reading the current frequency in Hz
        from rigctld. No special error handling is supplied.

        :return: None if error else  float frequency in Hz
        """
        result = None
        if self._socket:
            self._socket.sendall("f"+chr(13))
            data = self._socket.recv(1000)
            log(3, "hamlib_read_frequency: %s" % data)
            if data:
                if not data.startswith("RPRT"):
                    result = float(data.strip())
        return result

    def set_frequency(self, frequency):
        """
        sets the frequency of connected trx
        :return Boolean
        """
        result = False
        if self._socket:
            log(3, "hamlib_set_frequency: %s" % frequency)
            self._socket.sendall("F %d" % int(frequency) +  chr(13))
            data = self._socket.recv(1000)
            if data:
                result = data.startswith("RPRT 0")
        return result

    def set_split_frequency(self, frequency):
        """
        sets the frequency of connected trx
        :return Boolean
        """
        result = False
        if self._socket:
            log(3, "hamlib_set_split_frequency: %s" % frequency)
            self._socket.sendall("I %d" % int(frequency) +  chr(13))
            data = self._socket.recv(1000)
            if data:
                result = data.startswith("RPRT 0")
        return result

_interactive_commands = ["controlling offset frequency by:"
                         "  a = -1000Hz",
                         "  s =  -100Hz",
                         "  d =   -10Hz",
                         "  f =    -1Hz",
                         "  g =    +1Hz",
                         "  h =   +10Hz",
                         "  j =  +100Hz",
                         "  k = +1000Hz",
                         "  r = reset to 0Hz",
                         "b = calibrate tx by beacon offset with input",
                         "B = calibrate tx by beacon offset from gqrx",
                         "n = input of upconverteroffset in Hz to beacon frequency",
                         "Upper case sets split frequency offset",
                         "? for Help (this message)",
                         "t = sync from trx",
                         "p = pause sync",
                         "c = continue sync",
                         "0 = show last frequencies",
                         "1..9 = last frequencies",
                         "+/- <freq in khz> = relative change of frequency",
                         "! <cmd> : call shell command synchron", 
                         "q for exit"]


def interactive_commands():
    has_7log = os.path.isfile("/usr/bin/7log")
    cmds = []
    cmds.extend(_interactive_commands)
    if has_7log:
        cmds.append("!l call 7log")
    cmd_str = "\n".join(cmds)
    return cmd_str


class FreqHistory(object):

    def __init__(self, max_size=10, min_diff=3000):
        self.max_size = 10
        self.min_diff = min_diff
        self._data = []

    def add_frequency(self, f):
        """
        adds f to list, if f is member
        of the list, move f to front
        if f - data[0] < min_diff overwrite front
        """
        intf = int(f)
        add_new = False
        if  not self._data:
            add_new = True
        else:
            if abs(self._data[0] - f) > self.min_diff:
                add_new = True
            else:
                self._data[0] = f
        if add_new:
            i = -1
            try:
                i = self._data.index(f)
            except ValueError:
                pass
            if i >= 0:
                self._data.remove(f)
                self._data.insert(0, f)
            else:
                self._data.insert(0, f)
                if len(self._data) >= self.max_size:
                    self._data.pop()

    def get_frequency(self, ind):
        if ind > 0 and ind <= len(self._data):
            return self._data[ind - 1]
            
    def get_frequency_list(self):
        return self._data

class TrxTransfer(object):

    def __init__(self, 
                 src_host,
                 dst_host,
                 f_offset,
                 split_offset,
                 min_qrg,
                 max_qrg,
                 min_tx,
                 max_tx,
                 split_mode=True,
                 qoverbose=False):
        self.src_host = src_host
        self.dst_host = dst_host
        self.min_qrg = min_qrg
        self.max_qrg = max_qrg
        self.min_tx = min_tx
        self.max_tx = max_tx
        self.f_offset = f_offset
        self.split_offset = split_offset
        self.split_mode = split_mode
        self.show_qo100_verbose=qoverbose
        self.finish = False
        self._enable_sync = True
        self.msg_queue = collections.deque()
        self.thread = threading.Thread(target=self.transfer_frequency, args=())
        self.f_list = FreqHistory()

    def transfer_frequency(self):
        """
        Transfers frequency from qgrx to IC910H
        Sets RX und TX frequencies
        """
        global _initial_split_offset
        last_qrg = -1
        last_split_offset = 0
        while not self.finish:
            try:
                with HamLibConnection(self.dst_host) as trx:
                    with HamLibConnection(self.src_host) as gqrx:
                        while not self.finish:
                            time.sleep(0.5)
                            self.process_queue(gqrx, trx)
                            f = gqrx.read_frequency()
                            if self._enable_sync and f is not None:
                                dst_f = f + self.f_offset
                                # print "F: %s %s %s " % (dst_f, self.min_qrg, self.max_qrg)
                                if last_qrg !=  dst_f or last_split_offset != self.split_offset:
                                    last_qrg = dst_f
                                    last_split_offset = self.split_offset
                                    if dst_f >= self.min_qrg and dst_f <= self.max_qrg :
                                        if not trx.set_frequency(dst_f):
                                            show_error("Setting frequency failed")
                                        else:
                                            log(1, "New frequency: %f" % (dst_f / 1e6))
                                            if self.show_qo100_verbose and self.split_mode:
                                                show_qo100(f+self.split_offset - _initial_split_offset)
                                        self.f_list.add_frequency(dst_f)
                                        self.handle_split(dst_f, trx)
                                        #tx_f = dst_f + self.split_offset
                                        #if tx_f >= self.min_tx and tx_f < self.max_tx and self.split_mode:
                                        #    if not trx.set_split_frequency(dst_f + self.split_offset):
                                        #        show_error("Setting TX-Frequency failed")

                                    else:
                                        show_error("Frequency %f MHz out of range" % (dst_f / 1e6))
            except Exception as e:
                show_error("ConnectionError: %s\n"
                           "gqrx or hamlib not running." % e)
                time.sleep(5)

    def handle_split(self, dst_f, trx):
        tx_f = dst_f + self.split_offset
        if tx_f >= self.min_tx and tx_f < self.max_tx and self.split_mode:
            if not trx.set_split_frequency(dst_f + self.split_offset):
                show_error("Setting TX-Frequency failed")

    def add_queue(self, callable_func, args=None):
        """
        adds callable fiwth signare c(gqrx, trx) to the queue
        """
        self.msg_queue.append((callable_func,args))

    def process_queue(self, gqrx, trx):
        if len(self.msg_queue):
            c = self.msg_queue.pop()
            if c:
                func, args = c
                func(gqrx,trx,args)

    def enable_sync(self, enable):
        self._enable_sync = enable
  
    def correct_rx(self, gqrx, trx, args):
        """
        args = frequency offset to add add rx frequency
        """  
        f_gqrx = gqrx.read_frequency()
        f_trx = trx.read_frequency()
        if f_gqrx:
            f_gqrx += args
            gqrx.set_frequency(f_gqrx)
        if f_trx:
            f_trx += args
            if f_trx >= self.min_qrg and f_trx <= self.max_qrg :
                trx.set_frequency(f_trx)
        
    def sync_from_trx(self, gqrx, trx, args):
        f = trx.read_frequency()
        log(2, "Sync: New gqrx frequency: %d" % f)
        if f is not None:
            dst_f = f - self.f_offset
            log(1, "Sync: New gqrx frequency: %d" % dst_f)
            if not gqrx.set_frequency(dst_f):
                show_error("Can't set frequency %d in gqrx." % dst_f)
        else:
            show_error("Can't get frequency from trx")

    def set_offset_by_beacon(self, gqrx, trx, args):
        beacon = gqrx.read_frequency()
        if beacon:
            if beacon >= self.min_qrg and beacon <= self.max_qrg:
                new_trx = _initial_split_offset + (_target_beacon_freq - beacon) 
                self.set_split_offset(new_trx)
            else:
                six.print_("Beacon not in RX range")

    def set_history(self, gqrx, trx, args):
        f = self.f_list.get_frequency(args)
        if f:
            six.print_("Setting f to :%s" % f) 
            gqrx.set_frequency(f)
    
    def rel_f_change(self, gqrx, trx, args):
        f = gqrx.read_frequency()
        gqrx.set_frequency(f+args)
        
    def set_f_offset(self, offset):
        self.f_offset = offset

    def set_split_offset(self, offset):
        # wennn der differenz kleiner wird,
        # steigt die Sendefrquenz
        # damit muss  bei gleichbleibender
        # tx frequenz, die rx frequenz 
        # um den faktor kleiner werden.
        # d.h. empfangsfrquenzkorrekturen
        # muessen dann uber das Programm
        # gemacht werden, damit sich die
        # tx qrg nicht aendert
        # die tx qrg wird nach autmatisch
        # durch den sync danach wieder auf
        # die richtige frquenz gestellt
        diff = offset - self.split_offset
        self.split_offset = offset
        #if diff:
        #    self.add_queue(self.correct_rx, diff)

    def start(self):
        self.thread.start()

    def terminate(self):
        if not self.finish:
            self.finish = True
            self.thread.join()
            
    def get_current_frequency(self):
        global _target_beacon_freq
        frequencies = self.f_list.get_frequency_list()
        real_f = None
        uplink = None
        if frequencies:
            current_f = frequencies[0]
            _qo_downlink = 10489800000
            lo_all = _qo_downlink - _target_beacon_freq 
            total_f = current_f - _target_beacon_freq
            real_f = (_qo_downlink + total_f)
            uplink = real_f - 8089500000 # Offset tx
        return real_f, uplink


def run_cmdline():
    global _verbose_level
    global _version
    global _finish
    global _f_offset
    global _initial_split_offset
    ap = argparse.ArgumentParser(description="gqrx_settrx. Version: %s" % _version,
                                 epilog="""
                                    All frequencies must be entred in Hz as 
                                    integer or float values.
                                    Transceiver IC910H must be in SAT NORM mode,
                                    for setting the tx frequency correctly via
                                    the split commands (I) of rigctld.

                                    This program is mainly for working with P4/A
                                    QO-100. Transmitting on 432 MHz and receiving
                                    on 144.500 MHz.
                                    Use rigtcld -m 344 -r /dev/ttyUSB0 -s 9600 to connect to IC910H.

                                    (C) DG7BBP, 2019
                                    """)
    ap.add_argument("--gqrx", required=False, default="localhost:7356", help="gqrx host:port")
    ap.add_argument("--verbose", type=int, default=0, choices=range(0,4), help="Verbose message (3=debug, 1 Simple output)")
    ap.add_argument("--trx",
                    default="localhost:4532",
                    help="""rigctld- address <ip/hostname>[<:port>] of TRX
                            """)
    ap.add_argument("--splitmode", action='store_true', help="Set TX in splitmode (enables tx)")
    # later for better use extra txclient with out rx client. so we can use two one tx and one rx
    # both controlled by the src (gqrx)
    # ap.add_argument("--onlytx", action='store_true', help="Do not set RX only set TX")
    ap.add_argument("--min",
                    dest="minrx",
                    type=float,
                    default=144000000,
                    help="Minimum receive or main frequency in Hz Default: 144MHz")
    ap.add_argument("--max",
                dest="maxrx",
                type=float,
                default=146000000,
                help="Maximum receive or main frequency in Hz . Default: 146MHz")
    ap.add_argument("--mintx",
                    type=float,
                    default=432000000,
                    help="Minimum transmit frequency in split mode. Default: 432MHz")
    ap.add_argument("--maxtx",
                type=float,
                default=434000000,
                help="Maximum transmit frequency in split mode. Default: 434MHz")
    ap.add_argument("--mainoffset",
                    dest="rxoffset",
                    type=float,
                    default = -1,
                    help="Receive offset from gqrx to TRX. Default:0: Overwrites stored value")
    ap.add_argument("--txoffset",
                    type=float,
                    default = -1,
                    # default=287500000,
                    help="Receive to transmit offset in trx to trx. Default 287500000. Overwrites stored value")
    ap.add_argument("--targetbeacon",
                     type=int,
                     default=144800000,
                     help="Target beacon frequency")
    ap.add_argument("--config",
                    default="gqrxsettrx_default",
                    help="Offset config to use")
    ap.add_argument("--qo100",
                    action="store_true",
                    help="Show QO-100 NB-frequency and mode")
    args = ap.parse_args()
    _verbose_level = args.verbose
    if args.rxoffset != -1.0:
        _f_offset = int(args.rxoffset)
        f_offset = _f_offset
    if args.txoffset != -1.0:
        _initial_split_offset = int(args.txoffset)
        split_offset = _initial_split_offset
    conf_file = args.config + ".conf"
    r_f_offset, r_split_offset = read_config(conf_file)
    if args.rxoffset == -1.0:
        f_offset = r_f_offset
    if args.txoffset == -1.0:
        split_offset = r_split_offset
    _target_beacon_freq = args.targetbeacon
    log(3, "Target beacon is %s" % _target_beacon_freq)

    # src_host, dst_host, f_offset, split_offset, min_qrg, max_qrg, min_tx, max_tx
    trx_t = TrxTransfer(args.gqrx,
                        args.trx,
                        f_offset,
                        split_offset,
                        int(args.minrx),
                        int(args.maxrx),
                        int(args.mintx),
                        int(args.maxtx),
                        args.splitmode,
                        args.qo100)

    trx_t.start()
    try:
        manage_input(trx_t)
    except KeyboardInterrupt:
        pass
    finally:
        trx_t.terminate()
    save_config(trx_t, conf_file)
    return 0


def call_logbook(cmd, trx):
    log_cmd = "7log %s" %cmd 
    call_shell(log_cmd, trx)


def call_shell(cmd, trx):
    if cmd:
        f, uplink  = trx.get_current_frequency()
        if f is not None:
            f_str = "%s" % f
            uplink_str = "%s" % uplink
            env = dict()
            env.update(os.environ)
            env["DG7BBP_LOG_QRG"] = uplink_str
            env["DG7BBP_LOG_QRG_RX"] = f_str
            args = shlex.split(cmd.strip())
            # six.print_("f: %s cmd '%s'" % (f_str, args))
            try:
                p  = subprocess.Popen(args,env=env)
                p.wait()
            except Exception as e:
               print("Error starting program ")


def manage_input(trx):
    """
    Read input for setting the offest frequencies
    """
    global _finish
    global _initial_split_offset
    global _target_beacon_freq
    fmap = {"a": -1000,
            "s": -100,
            "d": -10,
            "f": -1,
            "g": 1,
            "h": 10,
            "j": 100,
            "k": 1000}
    log(1,interactive_commands())
    while not _finish:
        c = rinput("Command: ").strip()
        if c.islower():
            f = fmap.get(c)
            if f is not None:
                trx.set_f_offset(trx.f_offset + f)
                six.print_("New offset: %d Hz" % trx.f_offset)
        elif c.isupper():
            f = fmap.get(c.lower())
            if f is not None:
                trx.set_split_offset(trx.split_offset + f)
                six.print_("New split offset: %d Hz" % trx.split_offset)
        if c == 'q':
            six.print_("Exiting")
            _finish = True
        elif c == "r":
            trx.set_f_offset(0)
            six.print_("New offset: %d Hz" % trx.f_offset)
        elif c == "R":
            trx.set_split_offset(_initial_split_offset)
            six.print_("New split offset: %d Hz" % trx.split_offset)
        elif c == "b":
            b_str = rinput("Beacon frequency (MHz):")  
            beacon= None
            try:
                beacon = int(float(b_str)*1e6)
            except ValueError:
                show_error("Invalid input tx offset not set")
            if beacon is not None:
                if beacon >= trx.min_qrg and beacon <= trx.max_qrg:
                    if trx.split_mode:
                        new_trx = _initial_split_offset + (_target_beacon_freq - beacon) 
                        trx.set_split_offset(new_trx)
                    else:
                        new_rx_offset = _f_offset + (_target_beacon_freq - beacon)
                        trx.set_f_offset(new_rx_offset)
                else:
                    six.print_("Beacon not in RX range")
        elif c == "B":
            if trx.split_mode:
                trx.add_queue(trx.set_offset_by_beacon)
            else:
                six.print_("Set offset by gqrx frequency is only available in split mode")
        elif c == '?':
            six.print_(interactive_commands())
        elif c == 't':
            trx.add_queue(trx.sync_from_trx)
            six.print_("Sync to gqrx")
        elif c == "p":
            trx.enable_sync(False)
            six.print_("Pause press c to continue")
        elif c == "c":
            trx.enable_sync(True)
            six.print_("Pause disabled")
        elif c == "n":
            f_str = rinput("Absolut rx offset Hz:")
            f_offset = None
            try:
                f_offset=int(f_str)
            except ValueError:
                show_error("Invalid input rx_offset not set")
            if f_offset is not None:
                trx.set_f_offset(trx.f_offset + f)
                six.print_("New offset: %d Hz" % trx.f_offset)
        elif c.isdigit():
            if c == "0":
                i = 1
                for f in trx.f_list.get_frequency_list():
                    six.print_("%d: %d" % (i, f))
                    i += 1
            else:
                trx.add_queue(trx.set_history, int(c)) 
        elif c[0] in ["+" , "-"] and c[1:].isdigit():
            trx.add_queue(trx.rel_f_change, int(c) * 1000)
        elif c.startswith("!l ") or c == "!l":
            call_logbook(c[2:], trx)
        elif c[0] == "!":
            call_shell(c[1:], trx)


def config_file(config_name):
    """
    get config filename
    if running on windows try LOCALAPPDATA then APPDATA
    """
    config_file = None
    config_name = config_name.replace("/", "_").replace(">","_")
    if sys.platform == "win32":
        appdata = os.environ.get("LOCALAPPDATA")
        if not appdata:
            appdata =  os.environ.get("APPDATA")
        if appdata:
            config_dir = os.path.join(appdata, "dg7bbp", "gqrxsettrx")
            if not os.path.isdir(config_dir):
                os.makedirs(config_dir)
            config_file = os.path.join(config_dir, config_name)
    else:
        home_dir = os.environ.get("HOME")
        if home_dir:
            config_dir = os.path.join(home_dir, ".config", "gqrxsettrx")
            if not os.path.isdir(config_dir):
                os.makedirs(config_dir)
            config_file = os.path.join(config_dir, config_name)
    return config_file


def read_config(config_name):
    """
    Reading config data into global variables
    """
    global _initial_split_offset
    c_file = config_file(config_name)
    f_offset = 0
    split_offset = _initial_split_offset
    if os.path.isfile(c_file):
        with open(c_file) as f:
            data = json.load(f)
        if data and type(data) == dict and data.get("id") == "qqrxconfig":
            f_offset = data.get("f_offset", 0)
            log(1, "Offset frequency %d " % f_offset)
            split_offset = data.get("split_offset", _initial_split_offset)
            log(1, "Split Offset frequency %d " % split_offset)
    else:
        log(0, "Config file :%s not found" % config_name)
    return f_offset, split_offset


def save_config(trx, config_name):
    """
    Save config data from global variables
    """
    with open(config_file(config_name), "w") as f:
        data = {"id": "qqrxconfig",
                "f_offset": trx.f_offset,
                "split_offset": trx.split_offset}
        json.dump(data, f)

if __name__  == "__main__":
    run_cmdline()
